"""Functions to calculate training loss on batches of images.

While functions are comparable to the ones found in the module metrics,
these rely on keras' backend and do not take raw numpy as input.
"""

import torch
import torch.nn.functional as F
import torch.nn as nn
from monai.losses import DiceLoss
import warnings
import pdb

class CompoundLoss(nn.Module):
    def __init__(self, loss1, loss2=None, alpha1=1., alpha2=0.):
        super(CompoundLoss, self).__init__()
        self.loss1 = loss1
        self.loss2 = loss2
        self.alpha1 = alpha1
        self.alpha2 = alpha2

    def forward(self, y_pred, y_true, y_true_sk=None):
        if y_true_sk:
            l1 = self.loss1(y_pred, y_true, y_true_sk)
            l2 = self.loss2(y_pred, y_true, y_true_sk)
        else:
            l1 = self.loss1(y_pred, y_true)
            l2 = self.loss2(y_pred, y_true)
        if self.alpha2 == 0 or self.loss2 is None:
            return self.alpha1*l1
        return self.alpha1*l1 + self.alpha2 * l2

class DSCLoss(nn.Module):
    def __init__(self, include_background=False):
        super(DSCLoss, self).__init__()
        self.loss = DiceLoss(softmax=True, include_background=include_background)
        self.check_softmax = True

    def forward(self, y_pred, y_true):
        if self.check_softmax:
            if y_pred.softmax(dim=1).flatten(2).sum(dim=1).mean(dim=1).mean() != 1.0:
                # flatten(2) flattens all after dim=2, sum over classes, take mean to get per-batch values, & take mean
                warnings.warn('check you did not apply softmax before loss computation')
            else: self.check_softmax = False
        return self.loss(y_pred, y_true)

class CELoss(nn.Module):
    def __init__(self):
        super(CELoss, self).__init__()
        self.loss = torch.nn.CrossEntropyLoss()
    def forward(self, y_pred, y_true):
        return self.loss(y_pred, y_true)


def soft_erode(img):
    if len(img.shape) == 4:
        p1 = -F.max_pool2d(-img, (3, 1), (1, 1), (1, 0))
        p2 = -F.max_pool2d(-img, (1, 3), (1, 1), (0, 1))
        return torch.min(p1, p2)
    elif len(img.shape) == 5:
        p1 = -F.max_pool3d(-img, (3, 1, 1),(1, 1, 1), (1, 0, 0))
        p2 = -F.max_pool3d(-img, (1, 3, 1), (1, 1, 1), (0, 1, 0))
        p3 = -F.max_pool3d(-img, (1, 1, 3), (1, 1, 1), (0, 0, 1))
        return torch.min(torch.min(p1, p2), p3)


def soft_dilate(img):
    if len(img.shape) == 4:
        return F.max_pool2d(img, (3, 3), (1, 1), (1, 1))
    elif len(img.shape) == 5:
        return F.max_pool3d(img,(3, 3, 3),(1, 1, 1), (1, 1, 1))


def soft_open(img):
    return soft_dilate(soft_erode(img))


def soft_skel(img, iter_):
    img1 = soft_open(img)
    skel = F.relu(img-img1)
    for j in range(iter_):
        img = soft_erode(img)
        img1 = soft_open(img)
        delta = F.relu(img-img1)
        skel = skel + F.relu(delta-skel*delta)
    return skel


class clDiceLoss(torch.nn.Module):
    def __init__(self, iters=3, smooth=1., include_background=False):
        super(clDiceLoss, self).__init__()
        self.iters = iters
        self.smooth = smooth
        self.include_background = include_background
        self.check_softmax = True

    def forward(self, y_pred, y_true):
        y_pred = y_pred.softmax(dim=1)
        y_pred_sk = soft_skel(y_pred, self.iters)
        y_true_sk = soft_skel(y_true, self.iters)
        tprec = (torch.sum(torch.multiply(y_pred_sk, y_true)[:, 1, ...]) + self.smooth)/(torch.sum(y_pred_sk[:, 1, ...]) + self.smooth)
        tsens = (torch.sum(torch.multiply(y_true_sk, y_pred)[:, 1, ...]) + self.smooth)/(torch.sum(y_true_sk[:, 1, ...]) + self.smooth)
        cl_dice = 1. - 2.0*(tprec*tsens)/(tprec+tsens)
        return cl_dice


class clCELoss(torch.nn.Module):
    def __init__(self, iters=3, smooth=1., include_background=False):
        super(clCELoss, self).__init__()
        self.iters = iters
        self.smooth = smooth
        self.include_background = include_background

    def forward(self, y_pred, y_true):
        l_unred = torch.nn.functional.cross_entropy(y_pred, y_true, reduction='none')
        y_pred = y_pred.softmax(dim=1)
        y_pred_sk = soft_skel(y_pred, self.iters)
        y_true_sk = soft_skel(y_true, self.iters)

        tprec = torch.mul(l_unred, y_true_sk[:, 1]).mean()
        tsens = torch.mul(l_unred, y_pred_sk[:, 1]).mean()
        cl_ce = (tprec+tsens)
        return cl_ce


def get_loss(loss1, loss2=None, alpha1=1., alpha2=0.):
    if loss1 == loss2 and alpha2 != 0.:
        warnings.warn('using same loss twice, you sure?')
    loss_dict = dict()
    loss_dict['ce'] = CELoss()
    loss_dict['dice'] = DSCLoss()
    loss_dict['cedice'] = CompoundLoss(CELoss(), DSCLoss(), alpha1=1., alpha2=1.)
    loss_dict['cldice'] = clDiceLoss()
    loss_dict['clce'] = clCELoss()

    loss_dict[None] = None

    loss_fn = CompoundLoss(loss_dict[loss1], loss_dict[loss2], alpha1, alpha2)

    return loss_fn


import sys
from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget, QVBoxLayout, QLabel, QSlider, QHBoxLayout, QAction, QFileDialog, QToolTip
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPixmap, QImage
import tifffile as tiff  # For loading TIFF files


class VolumeViewer(QMainWindow):
    def __init__(self):
        super().__init__()
        self.volume = None
        self.filename = None
        self.current_slice = 0
        self.initUI()

    def initUI(self):
        self.setWindowTitle('Volume Viewer | Fluospotter')

        # Create menu bar
        menubar = self.menuBar()
        file_menu = menubar.addMenu('File')

        # Add actions to the menu
        load_action = QAction('Load Volume...', self)
        load_action.triggered.connect(self.load_volume)
        file_menu.addAction(load_action)
        file_menu.addSeparator()
        exit_action = QAction('Exit', self)
        exit_action.triggered.connect(self.close)
        file_menu.addAction(exit_action)

        # Main widget
        self.main_widget = QWidget()
        self.main_layout = QHBoxLayout()

        # Image display
        self.image_label = QLabel()
        self.main_layout.addWidget(self.image_label)

        # Control panel
        self.control_panel = QVBoxLayout()

        # File name label
        self.file_label = QLabel('File: ')
        self.control_panel.addWidget(self.file_label)

        # Slice slider
        slice_layout = QHBoxLayout()
        self.slice_label = QLabel('Slice: 0')
        slice_layout.addWidget(self.slice_label)
        self.slice_slider = QSlider(Qt.Horizontal)
        self.slice_slider.setMinimum(0)
        self.slice_slider.setMaximum(0)
        self.slice_slider.setValue(0)
        self.slice_slider.valueChanged.connect(self.change_slice)
        slice_layout.addWidget(self.slice_slider)

        self.control_panel.addLayout(slice_layout)
        self.main_layout.addLayout(self.control_panel)
        self.main_widget.setLayout(self.main_layout)
        self.setCentralWidget(self.main_widget)

        # Delete later #
        self.filename = r"C:\Users\arnau.blanco\OneDrive - Bruker Physik GmbH\Documents\GitHub\Fluospotter\data\test\Location-03.tif"
        self.volume = tiff.imread(self.filename)
        self.slice_slider.setMaximum(self.volume.shape[0] - 1)
        self.slice_slider.setValue(0)
        filename = self.filename
        if len(filename) > 30: filename = filename[:30] + "..."
        self.file_label.setText(f'File: {filename}')
        self.file_label.setToolTip(self.filename)  # Set tooltip with full file name
        self.update_image()
        ##

    def load_volume(self):
        options = QFileDialog.Options()
        options |= QFileDialog.ReadOnly
        file_name, _ = QFileDialog.getOpenFileName(self, "Open Volume File", "", "TIFF Files (*.tif);;All Files (*)",
                                                   options=options)

        if file_name:
            self.filename = file_name
            self.volume = tiff.imread(file_name)
            self.slice_slider.setMaximum(self.volume.shape[0] - 1)
            self.slice_slider.setValue(0)
            filename = self.filename
            if len(filename) > 30: filename = filename[:30] + "..."
            self.file_label.setText(f'File: {filename}')
            self.file_label.setToolTip(self.filename)  # Set tooltip with full file name
            self.update_image()

    def change_slice(self, value):
        self.current_slice = value
        self.slice_label.setText(f"Slice: {value}")
        self.update_image()

    def update_image(self):
        if self.volume is not None:
            slice_image = self.volume[self.current_slice]
            qimage = QImage(slice_image.data, slice_image.shape[1], slice_image.shape[0], QImage.Format_Grayscale16)
            pixmap = QPixmap.fromImage(qimage)
            self.image_label.setPixmap(pixmap)
            self.image_label.setFixedSize(pixmap.size())


def main():
    app = QApplication(sys.argv)
    viewer = VolumeViewer()
    viewer.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()


"""Dataset class."""

import os
from ..io import load_files, validate_data_and_labels


class Dataset:
    """Class for datasets.

    Args:
        data_dir (str): Absolute path to the data directory.
        segmentation_dir (str, optional): Absolute path to the segmentation labels directory.
        spots_dir (str, optional): Absolute path to the spots labels directory.
    """

    def __init__(self, data_dir: str, segmentation_dir: str = None, spots_dir: str = None, training: bool = False):
        self.data_dir = data_dir
        self.segmentation_dir = segmentation_dir
        self.spots_dir = spots_dir
        self.training = training
        self.load_data()

    def load_data(self) -> None:
        """Load dataset paths into memory and validate their consistency."""
        # Load the data files
        data = load_files(self.data_dir, self.training)
        if self.training:
            self.x_train, self.x_valid, self.x_test = data["train"], data["valid"], data["test"]
        else:
            self.x_test = data["test"]

        # Load and validate segmentation labels if available
        if self.segmentation_dir is not None:
            labels_seg = load_files(self.segmentation_dir, self.training)
            if not validate_data_and_labels(data, labels_seg, self.training):
                raise ValueError("Data files and segmentation annotations do not coincide.")
            if self.training:
                self.y_seg_train, self.y_seg_valid, self.y_seg_test = labels_seg["train"], labels_seg["valid"], labels_seg["test"]
            else:
                self.y_seg_test = labels_seg["test"]

        # Load and validate spots labels if available
        if self.spots_dir is not None:
            labels_spots = load_files(self.spots_dir, self.training)
            if not validate_data_and_labels(data, labels_spots, self.training):
                raise ValueError("Data files and spots annotations do not coincide.")
            if self.training:
                self.y_spots_train, self.y_spots_valid, self.y_spots_test = labels_spots["train"], labels_spots["valid"], labels_spots["test"]
            else:
                self.y_spots_test = labels_spots["test"]

        # Prepare and normalize the data
        self.prepare_data()
        self.normalize_dataset()

    def segmentation_data_test(self):
        return self.x_test, self.y_seg_test

    def segmentation_data_train(self):
        return self.x_train, self.y_seg_train

    def segmentation_data_valid(self):
        return self.x_valid, self.y_seg_valid

    def spots_data_test(self):
        return self.x_test, self.y_spots_test

    def spots_data_train(self):
        return self.x_train, self.y_spots_train

    def spots_data_valid(self):
        return self.x_valid, self.y_spots_valid

    def prepare_data(self):
        """Empty method to prepare or convert data."""

    def normalize_dataset(self):
        """Empty method to normalise images in the dataset."""



"""Model class, to be extended by specific types of models."""

from typing import List
import numpy as np
from ..datasets import Dataset


class Model:
    """Base class to be subclassed by predictors for specific types of data."""

    def __init__(self) -> None:
        pass

    def train(self, dataset: Dataset) -> None:
        """Training loop."""
        pass

    def evaluate(self, x: np.ndarray, y: np.ndarray) -> List[float]:
        """Evaluation function."""
        pass


"""List of functions to handle data including converting matrices <-> coordinates."""
import pdb
from typing import Tuple, Dict
import os
import numpy as np
import torch
import monai.data as md
from torch.utils.data.dataset import Subset
from .augment import get_transforms_fullres, get_transforms_patches
from .metrics import iou


def get_test_split(data_path='data', labels_path=''):
    data_test = os.path.join(data_path, 'test')
    vol_list_test = os.listdir(data_test)
    vol_list_test = [os.path.join(data_test, n) for n in vol_list_test]

    if len(labels_path) > 0:
        annotations_test = os.path.join(labels_path, 'test')
        seg_list_test = os.listdir(annotations_test)
        seg_list_test = [os.path.join(annotations_test, n) for n in seg_list_test]
        test_files = [{'img': img, 'seg': seg} for img, seg in zip(vol_list_test, seg_list_test)]
    else:
        test_files = [{'img': img, 'seg': ''} for img in zip(vol_list_test)]

    return test_files


def get_train_val_test_splits(data_path='data', labels_path='labels'):
    data_train, data_valid = os.path.join(data_path, 'train'), os.path.join(data_path, 'valid')
    annotations_train, annotations_valid = os.path.join(labels_path, 'train'), os.path.join(labels_path, 'valid')

    vol_list_train, vol_list_valid = os.listdir(data_train), os.listdir(data_valid)
    seg_list_train, seg_list_valid = os.listdir(annotations_train), os.listdir(annotations_valid)
    vol_list_train, vol_list_valid = [os.path.join(data_train, n) for n in vol_list_train], [os.path.join(data_valid, n) for n in vol_list_valid]
    seg_list_train, seg_list_valid = [os.path.join(annotations_train, n) for n in seg_list_train], [os.path.join(annotations_valid, n) for n in seg_list_valid]
    tr_files, vl_files = [{'img': img, 'seg': seg} for img, seg in zip(vol_list_train, seg_list_train)], [{'img': img, 'seg': seg} for img, seg in zip(vol_list_valid, seg_list_valid)]
    print(40*'=')
    print('* Total Number of Volumes {}'.format(len(tr_files)+len(vl_files)))
    print(40 * '=')

    print('* Training samples = {}, Validation samples = {}'.format(len(tr_files), len(vl_files)))

    return tr_files, vl_files


def get_loaders_fullres(data_path, batch_size=1, im_size=(96, 96, 64), num_workers=0, tr_percentage=1., ovft_check=0):

    tr_files, vl_files = get_train_val_test_splits(data_path)

    if tr_percentage < 1.:
        print(40 * '-')
        n_tr_examples = len(tr_files)
        random_indexes = np.random.permutation(n_tr_examples)
        kept_indexes = int(n_tr_examples * tr_percentage)
        tr_files = [tr_files[i] for i in random_indexes[:kept_indexes]]
        print('Reducing training data from {} items to {}'.format(n_tr_examples, len(tr_files)))
        print(40 * '-')

    tr_transforms, vl_transforms = get_transforms_fullres(im_size=im_size)

    test_batch_size = 2*batch_size
    gpu = torch.cuda.is_available()
    tr_ds = md.Dataset(data=tr_files, transform=tr_transforms)

    vl_ds = md.Dataset(data=vl_files, transform=vl_transforms)
    tr_loader = md.DataLoader(tr_ds, batch_size=batch_size, num_workers=num_workers, shuffle=True, pin_memory=gpu)
    vl_loader = md.DataLoader(vl_ds, batch_size=test_batch_size, num_workers=num_workers, pin_memory=gpu)
    if ovft_check > 0:
        ovft_ds = md.Dataset(data=tr_files, transform=vl_transforms)
        subset_size = len(vl_ds)
        subset_idxs = torch.randperm(len(ovft_ds))[:subset_size]
        ovft_ds = Subset(ovft_ds, subset_idxs)
    else: ovft_ds = md.Dataset(data=tr_files, transform=vl_transforms)
    ovft_loader = md.DataLoader(ovft_ds, batch_size=test_batch_size, num_workers=num_workers, pin_memory=gpu)

    return tr_loader, ovft_loader, vl_loader


def get_loaders(data_path, labels_path, n_samples=1, neg_samples=1, patch_size=(48, 256, 256), num_workers=0, ovft_check=0, depth_last=False, n_classes=2):

    tr_files, vl_files = get_train_val_test_splits(data_path, labels_path)
    tr_transforms, vl_transforms = get_transforms_patches(n_samples, neg_samples, patch_size=patch_size, depth_last=depth_last, n_classes=n_classes)
    batch_size = 1
    test_batch_size = 1
    gpu = torch.cuda.is_available()

    tr_ds = md.Dataset(data=tr_files, transform=tr_transforms)
    vl_ds = md.Dataset(data=vl_files, transform=vl_transforms)
    tr_loader = md.DataLoader(tr_ds, batch_size=batch_size, num_workers=num_workers, shuffle=True, pin_memory=gpu)
    vl_loader = md.DataLoader(vl_ds, batch_size=test_batch_size, num_workers=0, pin_memory=gpu)
    
    if ovft_check > 0:
        ovft_ds = md.Dataset(data=tr_files, transform=vl_transforms)
        subset_size = len(vl_ds)
        subset_idxs = torch.randperm(len(ovft_ds))[:subset_size]
        ovft_ds = Subset(ovft_ds, subset_idxs)
    else: ovft_ds = md.Dataset(data=tr_files, transform=vl_transforms)
    ovft_loader = md.DataLoader(ovft_ds, batch_size=test_batch_size, num_workers=0, pin_memory=gpu)
    return tr_loader, ovft_loader, vl_loader


def get_loaders_test(data_path, labels_path, n_samples=1, neg_samples=1, patch_size=(48, 256, 256), num_workers=0, depth_last=False, n_classes=2, im_size=(48, 512, 512), instance_seg=False):
    test_files = get_test_split(data_path, labels_path)
    _, test_transforms = get_transforms_patches(n_samples, neg_samples, patch_size=patch_size,
                                                          depth_last=depth_last, n_classes=n_classes, im_size=im_size, instance_seg=instance_seg)
    batch_size = 1
    gpu = torch.cuda.is_available()
    test_ds = md.Dataset(data=test_files, transform=test_transforms)
    test_loader = md.DataLoader(test_ds, batch_size=batch_size, num_workers=num_workers, shuffle=True, pin_memory=gpu)
    return test_loader


def join_connected_puncta(data, mask):
    out = np.zeros_like(mask, dtype=np.int8)
    labels = np.unique(mask)
    for l in labels[1:]:  # Skip label 0 (background)
        mask_label = mask == l
        if not np.any(mask_label):  # Skip empty masks
            continue
        # Find the index of the maximum value in the component
        max_idx = np.argmax(data[mask_label])
        # Create a mask with the maximum value position set to True
        max_pos_mask = np.zeros_like(mask_label)
        max_pos_mask[np.nonzero(mask_label)[0][max_idx],
                     np.nonzero(mask_label)[1][max_idx],
                     np.nonzero(mask_label)[2][max_idx]] = True
        out[max_pos_mask] = 1  # Set this position to 1 in the output mask
    return out


def match_labeling(actual, predicted):
    """
    Relabels the predicted segmentation labels to match the ground truth labels
    based on the highest Intersection over Union (IoU) overlap.

    Args:
        actual (np.array): Ground truth labels.
        predicted (np.array): Predicted labels.

    Returns:
        np.array: Relabeled predicted labels.
    """
    new_labels = np.zeros_like(predicted)
    counter = 0

    # Iterate over each unique predicted label (ignoring the background label 0)
    print("Total predicted labels:",len(np.unique(predicted))-1, "Total real labels:", len(np.unique(actual))-1)
    for pred_label in np.unique(predicted):
        print("Label:",pred_label)
        if pred_label == 0:
            continue
        current = (predicted == pred_label)  # Binary mask of current predicted label
        best_label, best_overlap = None, 0  # Initialize with no best label and best overlap

        # Iterate over each unique actual label (ignoring the background label 0)
        for actual_label in np.unique(actual):
            if actual_label == 0:
                continue
            overlap = iou(actual == actual_label, current)  # Compute IoU
            if overlap > best_overlap:  # If better overlap is found
                best_overlap = overlap
                best_label = actual_label

        # If a matching label is found, use it; otherwise, assign a new unique label
        if best_label is not None:
            new_labels[predicted == pred_label] = best_label
        else:
            # Assign a new unique label that is not in used_labels
            new_label = max(np.unique(actual)) + 1 + counter
            counter += 1
            new_labels[predicted == pred_label] = new_label

    return new_labels


"""Model utility functions for augmentation."""

import numpy as np
import torch
import monai.transforms as t
from tifffile import imread


def permute_depth(x):
    return torch.permute(x, [0, 2, 3, 1])


def custom_transform(d):
    img = torch.as_tensor(imread(d['img']).astype(np.float32)).unsqueeze(0)
    if 'seg' in d and d['seg']:
        seg = torch.as_tensor(imread(d['seg']).astype(np.int8)).unsqueeze(0)
        return {'img': img, 'seg': seg}
    else:
        return {'img': img}


def get_transforms_patches(n_samples, neg_samples, patch_size, im_size, n_classes=2, depth_last=False, p_app=0.1, pr_geom=0.1, instance_seg=False):
    def add_custom_layers(transforms_list):
        if not instance_seg:
            transforms_list.append(t.AsDiscreted(keys=('seg'), to_onehot=n_classes, allow_missing_keys=True))
        if depth_last:
            transforms_list.insert(1, t.Lambda(lambda d: {'img': permute_depth(d['img']), 'seg': permute_depth(d['seg'])}))

    tr_transforms = [
        t.Lambda(lambda d: {'img': torch.as_tensor(imread(d['img']).astype(np.float32)).unsqueeze(0),
                            'seg': torch.as_tensor(imread(d['seg']).astype(np.int8)).unsqueeze(0)}),
        t.ScaleIntensityd(keys=('img',)),
        t.Resized(spatial_size=im_size, keys=('img', 'seg'), mode=('bilinear', 'nearest')),
        t.RandCropByPosNegLabeld(keys=('img', 'seg'), label_key='seg', spatial_size=patch_size,
                                 num_samples=n_samples, pos=1, neg=neg_samples),
        t.RandScaleIntensityd(keys=('img',), factors=0.05, prob=p_app),
        t.RandShiftIntensityd(keys=('img',), offsets=0.05, prob=p_app),
        t.RandFlipd(keys=('img', 'seg'), prob=pr_geom, spatial_axis=0),
        t.RandFlipd(keys=('img', 'seg'), prob=pr_geom, spatial_axis=1),
        t.RandFlipd(keys=('img', 'seg'), prob=pr_geom, spatial_axis=2)
    ]

    vl_transforms = [
        t.Lambda(custom_transform),
        t.ScaleIntensityd(keys=('img',)),
        t.Resized(spatial_size=im_size, keys=('img', 'seg'), mode=('bilinear', 'nearest'), allow_missing_keys=True)
    ]

    # Apply depth permutation if needed
    add_custom_layers(tr_transforms)
    add_custom_layers(vl_transforms)

    return t.Compose(tr_transforms), t.Compose(vl_transforms)


def get_transforms_fullres(n_samples, neg_samples, patch_size, n_classes=2, depth_last=False, p_app=0.1, pr_geom=0.1, instance_seg=False):
    def add_custom_layers(transforms_list):
        if not instance_seg:
            transforms_list.append(t.AsDiscreted(keys=('seg'), to_onehot=n_classes))
        if depth_last:
            transforms_list.insert(1,
                                   t.Lambda(lambda d: {'img': permute_depth(d['img']), 'seg': permute_depth(d['seg'])}))
    tr_transforms = [
        t.Lambda(lambda d: {'img': torch.as_tensor(imread(d['img']).astype(np.float32)).unsqueeze(0),
                            'seg': torch.as_tensor(imread(d['seg']).astype(np.int8)).unsqueeze(0)}),
        t.CropForegroundd(keys=('img', 'seg'), source_key='seg', allow_smaller=False),
        t.ScaleIntensityd(keys=('img',)),
        t.RandCropByPosNegLabeld(keys=('img', 'seg'), label_key='seg', spatial_size=patch_size,
                                 num_samples=n_samples, pos=1, neg=neg_samples),
        t.RandScaleIntensityd(keys=('img',), factors=0.05, prob=p_app),
        t.RandShiftIntensityd(keys=('img',), offsets=0.05, prob=p_app),
        t.RandFlipd(keys=('img', 'seg'), prob=pr_geom, spatial_axis=0),
        t.RandFlipd(keys=('img', 'seg'), prob=pr_geom, spatial_axis=1),
        t.RandFlipd(keys=('img', 'seg'), prob=pr_geom, spatial_axis=2)
    ]

    vl_transforms = [
        t.Lambda(lambda d: {'img': torch.as_tensor(imread(d['img']).astype(np.float32)).unsqueeze(0),
                            'seg': torch.as_tensor(imread(d['seg']).astype(np.int8)).unsqueeze(0)}),
        t.CropForegroundd(keys=('img', 'seg'), source_key='img', allow_smaller=False),
        t.ScaleIntensityd(keys=('img',)),
    ]

    # Apply depth permutation if needed
    add_custom_layers(tr_transforms)
    add_custom_layers(vl_transforms)

    return t.Compose(tr_transforms), t.Compose(vl_transforms)



"""Utility helper functions."""

from typing import Callable, Iterable, Tuple, Union
import importlib
import os
import random

from PIL import Image
from PIL import UnidentifiedImageError
from PIL.TiffTags import TAGS
import numpy as np
import pandas as pd
import psutil
import threading
import time
import sys


def monitor_ram_usage(stop_event):
    # Get the virtual memory details
    mem = psutil.virtual_memory()
    total_ram_mb = mem.total / (1024 * 1024)  # Convert bytes to MB
    threshold_mb = 0.98 * total_ram_mb

    while not stop_event.is_set():
        # Get the current process
        process = psutil.Process(os.getpid())

        # Get the memory info
        mem_info = process.memory_info()

        # Convert bytes to MB
        vms_mb = mem_info.vms / (1024 * 1024)  # Virtual Memory Size

        # Check if the RAM usage exceeds the threshold
        if vms_mb > threshold_mb:
            print(f"Memory RAM exceeded {threshold_mb:.2f} MB, stopping the program.")
            sys.exit(1)

        # Sleep for a while before checking again
        time.sleep(1)



"""Training functions."""
import pdb
from typing import Dict
import datetime
import os, os.path
import torch
from tqdm import trange
from .data import get_loaders
from .optimizers import get_optimizer, get_scheduler
from .losses import get_loss
from sklearn.neighbors import KNeighborsClassifier
from .inference import validate
import numpy as np
import random
import time
from .models import Model
from .datasets import Dataset


def init_tr_info():
    tr_info = dict()
    tr_info['tr_dscs'], tr_info['vl_dscs'] = [], []
    tr_info['tr_aucs'], tr_info['vl_aucs'] = [], []
    tr_info['tr_losses'], tr_info['vl_losses'] = [], []

    return tr_info


def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group['lr']


def train_one_epoch(model, tr_loader, bs, acc_grad, loss_fn, optimizer, scheduler):
    model.train()
    device = 'cuda' if next(model.parameters()).is_cuda else 'cpu'
    n_opt_iters = 0
    with trange(len(tr_loader)) as t:
        step, n_elems, running_loss = 0, 0, 0
        for (i_batch, batch_data) in enumerate(tr_loader):
            n_samples = len(batch_data['seg'])
            for m in range(0, n_samples, bs):
                step += bs
                inputs, labels = (batch_data['img'][m:(m+bs)].to(device), batch_data['seg'][m:(m+bs)].to(device))
                outputs = model(inputs)
                loss = loss_fn(outputs, labels)
                loss = loss / acc_grad
                loss.backward()
                if ((n_opt_iters + 1) % acc_grad == 0) or (n_opt_iters + 1 == len(tr_loader)):
                    optimizer.step()
                    optimizer.zero_grad()
                n_opt_iters += 1
                lr = get_lr(optimizer)
                scheduler.step()
                optimizer.zero_grad()
                running_loss += loss.detach().item() * inputs.shape[0]
                n_elems += inputs.shape[0]
                run_loss = running_loss / n_elems

            t.set_postfix(LOSS_lr="{:.4f}/{:.6f}".format(run_loss, lr))
            t.update()


def set_tr_info(tr_info, epoch=0, ovft_metrics=None, vl_metrics=None, best_epoch=False):
    if best_epoch:
        tr_info['best_tr_dsc'] = tr_info['tr_dscs'][-1]
        tr_info['best_vl_dsc'] = tr_info['vl_dscs'][-1]
        tr_info['best_tr_auc'] = tr_info['tr_aucs'][-1]
        tr_info['best_vl_auc'] = tr_info['vl_aucs'][-1]
        tr_info['best_tr_loss'] = tr_info['tr_losses'][-1]
        tr_info['best_vl_loss'] = tr_info['vl_losses'][-1]
        tr_info['best_epoch'] = epoch
    else:
        tr_info['tr_dscs'].append(ovft_metrics[0])
        tr_info['vl_dscs'].append(vl_metrics[0])
        tr_info['tr_aucs'].append(ovft_metrics[1])
        tr_info['vl_aucs'].append(vl_metrics[1])
        tr_info['tr_losses'].append(ovft_metrics[-1])
        tr_info['vl_losses'].append(vl_metrics[-1])

    return tr_info


def start_training(model, optimizer, acc_grad, loss_fn, bs, tr_loader, ovft_loader, vl_loader, scheduler, metric, n_epochs, vl_interval, save_path):
    best_metric, best_epoch = -1, 0
    tr_info = init_tr_info()
    for epoch in range(n_epochs):
        print('Epoch {:d}/{:d}'.format(epoch + 1, n_epochs))
        # train one cycle
        train_one_epoch(model, tr_loader, bs, acc_grad, loss_fn, optimizer, scheduler)
        if (epoch + 1) % vl_interval == 0:
            with torch.inference_mode():
                ovft_metrics = validate(model, ovft_loader, loss_fn)
                vl_metrics = validate(model, vl_loader, loss_fn)
            tr_info = set_tr_info(tr_info, epoch, ovft_metrics, vl_metrics)
            s = get_eval_string(tr_info, epoch)
            print(s)
            with open(os.path.join(save_path, 'train_log.txt'), 'a') as f: print(s, file=f)
            # check if performance was better than anyone before and checkpoint if so
            if metric == 'DSC': curr_metric = tr_info['vl_dscs'][-1]
            elif metric == 'AUC': curr_metric = tr_info['vl_aucs'][-1]

            if curr_metric > best_metric:
                print('-------- Best {} attained. {:.2f} --> {:.2f} --------'.format(metric, best_metric, curr_metric))
                torch.save(model.state_dict(), os.path.join(save_path, 'best_model.pth'))
                best_metric, best_epoch = curr_metric, epoch + 1
                tr_info = set_tr_info(tr_info, epoch+1, best_epoch=True)
            else:
                print('-------- Best {} so far {:.2f} at epoch {:d} --------'.format(metric, best_metric, best_epoch))
    torch.save(model.state_dict(), os.path.join(save_path, 'last_model.pth'))
    del model, tr_loader, vl_loader
    torch.cuda.empty_cache()
    return tr_info


def train_model(
    model: Model,
    dataset: Dataset,
    refinement: bool = False
) -> None:
    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")

    # Reproducibility
    seed_value = 0
    set_seeds(seed_value, use_cuda)

    # Logging
    cfg = model.cfg
    if "save_path" in cfg:
        save_path = os.path.join(cfg["save_path"], 'experiments')
    else:
        save_path = os.path.join(dataset.data_dir, '../experiments')

    counter = 1
    if cfg["model_type"] == "segmentation":
        save_path = os.path.join(save_path, 'segmentation')
        while os.path.exists(os.path.join(save_path, 'experiment' + str(counter))):
            counter += 1
        save_path = os.path.join(save_path, 'experiment' + str(counter))
        labels_path = dataset.segmentation_dir
    elif cfg["model_type"] == "puncta_detection":
        save_path = os.path.join(save_path, 'puncta_detection')
        while os.path.exists(os.path.join(save_path, 'experiment' + str(counter))):
            counter += 1
        save_path = os.path.join(save_path, 'experiment' + str(counter))
        labels_path = dataset.spots_dir
    else:
        raise ValueError("Model type {} not recognized".format(cfg["model_type"]))

    os.makedirs(save_path, exist_ok=True)

    print('* Instantiating a {} model for {}'.format(model.model_name, cfg["model_type"]))
    in_c = 1
    n_classes = int(cfg["n_classes"])
    if refinement:
        model = model.refinement
        n_classes = 2
    else:
        model = model.network
    model = model.to(device)
    print('* Creating Dataloaders, batch size = {}, samples/vol = {}, workers = {}'.format(cfg["batch_size"], cfg["n_samples"], cfg["num_workers"]))

    tr_loader, ovft_loader, vl_loader = get_loaders(data_path=dataset.data_dir, labels_path=labels_path, n_samples=int(cfg["n_samples"]), neg_samples=int(cfg["neg_samples"]),
                                                    patch_size=tuple(map(int, cfg["patch_size"].split('/'))), num_workers=int(cfg["num_workers"]), ovft_check=int(cfg["ovft_check"]), depth_last=bool(cfg["depth_last"]),
                                                    n_classes=n_classes)

    print("Total params: {0:,}".format(sum(p.numel() for p in model.parameters() if p.requires_grad)))
    opt_cfg = {}
    if "lr" in cfg:
        opt_cfg["learning_rate"] = float(cfg["lr"])
    if "weight_decay" in cfg:
        opt_cfg["weight_decay"] = float(cfg["weight_decay"])
    if "momentum" in cfg:
        opt_cfg["momentum"] = float(cfg["momentum"])

    optimizer = get_optimizer(cfg["optimizer"], opt_cfg, model.parameters())
    if cfg["cyclical_lr"]:
        scheduler_name = 'cosineAnnealingWarmRestarts'
        T = int(cfg["vl_interval"]) * len(tr_loader) * int(cfg["n_samples"]) // int(cfg["batch_size"])
    else:
        scheduler_name = 'cosineAnnealingLR'
        T = int(cfg["n_epochs"]) * len(tr_loader) * int(cfg["n_samples"]) // int(cfg["batch_size"])

    scheduler = get_scheduler(scheduler=scheduler_name, optimizer=optimizer, T=T, eta_min=0)
    loss_fn = get_loss(cfg["loss1"], cfg["loss2"], float(cfg["alpha1"]), float(cfg["alpha2"]))

    print('* Instantiating loss function {:.2f}*{} + {:.2f}*{}'.format(cfg["alpha1"], cfg["loss1"], cfg["alpha2"],
                                                                       cfg["loss2"]))
    print('* Starting to train\n', '-' * 10)
    start = time.time()

    tr_info = start_training(model, optimizer, cfg["acc_grad"], loss_fn, int(cfg["batch_size"]), tr_loader, ovft_loader, vl_loader, scheduler,
                          cfg["metric"], int(cfg["n_epochs"]), int(cfg["vl_interval"]), save_path)
    end = time.time()

    hours, rem = divmod(end - start, 3600)
    minutes, seconds = divmod(rem, 60)
    print('Training time: {:0>2}h {:0>2}min {:05.2f}secs'.format(int(hours), int(minutes), seconds))

    with (open(os.path.join(save_path, 'log.txt'), 'a') as f):
        print(
            'Best epoch = {}/{}: Tr/Vl DSC = {:.2f}/{:.2f} - Tr/Vl AUC = {:.2f}/{:.2f} - Tr/Vl LOSS = {:.4f}/{:.4f}\n'.format(
                tr_info['best_epoch'], cfg["n_epochs"],
                tr_info['best_tr_dsc'], tr_info['best_vl_dsc'], tr_info['best_tr_auc'], tr_info['best_vl_auc'],
                tr_info['best_tr_loss'], tr_info['best_vl_loss']), file=f)
        for j in range(cfg["n_epochs"] // cfg["vl_interval"]):
            s = get_eval_string(tr_info, epoch=j, finished=True, vl_interval=cfg["vl_interval"])
            print(s, file=f)
        print('\nTraining time: {:0>2}h {:0>2}min {:05.2f}secs'.format(int(hours), int(minutes), seconds), file=f)

    print('Done. Training time: {:0>2}h {:0>2}min {:05.2f}secs'.format(int(hours), int(minutes), seconds))

    print('Finished.')


def get_eval_string(tr_info, epoch, finished=False, vl_interval=1):
    ep_idx = len(tr_info['tr_dscs'])-1
    if finished:
        ep_idx = epoch
        epoch = (epoch+1) * vl_interval - 1

    s = 'Ep. {}: Train||Val DSC: {:5.2f}||{:5.2f} - AUC: {:5.2f}||{:5.2f} - Loss: {:.4f}||{:.4f}'.format(
        str(epoch+1).zfill(3), tr_info['tr_dscs'][ep_idx], tr_info['vl_dscs'][ep_idx],
              tr_info['tr_aucs'][ep_idx], tr_info['vl_aucs'][ep_idx],
              tr_info['tr_losses'][ep_idx], tr_info['vl_losses'][ep_idx])
    return s


def set_seeds(seed_value, use_cuda):
    np.random.seed(seed_value)  # cpu vars
    torch.manual_seed(seed_value)  # cpu  vars
    random.seed(seed_value)  # Python
    if use_cuda:
        torch.cuda.manual_seed(seed_value)
        torch.cuda.manual_seed_all(seed_value)  # gpu vars
        torch.backends.cudnn.deterministic = True  # needed
        torch.backends.cudnn.benchmark = False


"""Resnet architecture."""

import math

import tensorflow as tf

from ._networks import OPTIONS_CONV
from ._networks import conv_block
from ._networks import inception_block
from ._networks import residual_block
from ._networks import squeeze_block
from ._networks import upconv_block


def __block(inputs, filters, block, l2):
    opts_conv = OPTIONS_CONV
    opts_conv["kernel_regularizer"] = tf.keras.regularizers.l2(l2) if l2 else None
    opts_conv["bias_regularizer"] = tf.keras.regularizers.l2(l2) if l2 else None
    if block == "convolutional":
        x = conv_block(inputs=inputs, filters=filters, n_convs=3, opts_conv=opts_conv)
    if block == "inception":
        x = inception_block(inputs=inputs, filters=filters, l2_regularizer=l2)
    if block == "residual":
        x = residual_block(inputs=inputs, filters=filters, opts_conv=opts_conv)
    x = squeeze_block(x=x)
    return x


def __encoder(inputs, filters, block, l2, dropout):
    x = __block(inputs, filters, block, l2)
    skip = tf.keras.layers.SpatialDropout2D(dropout)(x)
    x = tf.keras.layers.MaxPool2D(pool_size=(2, 2))(skip)
    return x, skip


def __decoder(inputs, skip, filters, block, l2):
    x = __block(inputs, filters, block, l2)
    x = upconv_block(inputs=x, skip=skip)
    return x


def unet(
    dropout: float = 0.2,
    cell_size: int = 4,
    filters: int = 5,
    ndown: int = 2,
    l2: float = 1e-6,
    block: str = "convolutional",
) -> tf.keras.models.Model:
    """Unet model with second, cell size dependent encoder.

    Note that "convolution" is the currently best block.

    Arguments:
        dropout: Percentage of dropout before each MaxPooling step.
        cell_size: Size of one cell in the prediction matrix.
        filters: Log_2 number of filters in the first inception block.
        ndown: Downsampling steps in the first encoder / decoder.
        l2: L2 value for kernel and bias regularization.
        block: Type of block in each layer. [options: convolutional, inception, residual]
    """
    if not math.log(cell_size, 2).is_integer():
        raise ValueError(f"cell_size must be a power of 2, but is {cell_size}.")

    # Input
    inputs = tf.keras.layers.Input(shape=(None, None, 1))
    x = inputs
    skip_layers = []

    # Encoder v1
    for n in range(ndown):
        x, skip = __encoder(x, 2 ** (filters + n), block, l2, dropout)
        skip_layers.append(skip)
    skip_bottom = x

    # Decoder
    for n, skip in enumerate(reversed(skip_layers)):
        x = __decoder(x, skip, 2 ** (filters + (ndown - n)), block, l2)

    # Encoder v2
    ndown_cell = int(math.log(cell_size, 2))
    for n in range(ndown_cell):
        x, _ = __encoder(x, 2 ** (filters + n), block, l2, dropout)

    # Logit
    if ndown == 2 and cell_size == 4:
        x = tf.keras.layers.Concatenate()([skip_bottom, x])
    x = __block(x, 2 ** (filters + ndown_cell), block, l2)
    x = tf.keras.layers.Conv2D(filters=3, kernel_size=1, strides=1)(x)
    x = tf.keras.layers.Activation("sigmoid")(x)
    model = tf.keras.Model(inputs=inputs, outputs=x)

    return model


"""Functions to calculate training loss on single image."""
import pdb
from typing import List, Optional, Tuple, Union
import warnings

import numpy as np
from scipy.spatial import cKDTree
from scipy.stats import rankdata
from typing import Dict

EPS = 1e-12


def compute_puncta_metrics(predicted: np.array, actual: np.array, metrics: Dict) -> Dict:
    if len(metrics) == 0:
        metrics = {
            'precision': [],
            'recall': [],
            'f1-score': [],
            'number_peaks': [],
            'ground_truth': []
        }
    predicted, actual = mask_to_coords_conversion(predicted), mask_to_coords_conversion(actual)
    precision, recall, f1_score_metric = precision_recall_f1_score_coordinates(actual, predicted)
    metrics['precision'].append(precision), metrics['recall'].append(recall), metrics['f1-score'].append(f1_score_metric)
    metrics['number_peaks'].append(len(predicted))
    metrics['ground_truth'].append(len(actual))

    return metrics


def compute_segmentation_metrics(predicted: np.array, actual: np.array, metrics: Dict) -> Dict:
    if len(metrics) == 0:
        metrics = {'pixel-wise': {
            'iou': [],
            'dice': [],
            'precision': [],
            'recall': [],
            'specificity': [],
            "f1-score": [],
            'roc_auc': [],
            'tpr': [],
            'fpr': [],
        }, 'object-wise': {
            'detection_accuracy': [],
            'segmentation_quality': [],
            'panoptic_quality': [],
            'precision': [],
            'recall': [],
            'specificity': [],
            "f1-score": [],
            'roc_auc': [],
            'tpr': [],
            'fpr': []
        }}
    matches = matched_segments(actual, predicted)
    pq, da, sq = panoptic_quality(actual, predicted, matches)
    metrics['object-wise']['detection_accuracy'].append(da)
    metrics['object-wise']['segmentation_quality'].append(sq)
    metrics['object-wise']['panoptic_quality'].append(pq)
    precision, recall, f1_score_metric = precision_recall_f1_score(matches)
    metrics['object-wise']['precision'].append(precision), metrics['object-wise']['recall'].append(recall)
    metrics['object-wise']['f1-score'].append(f1_score_metric)

    actual_bin, predicted_bin = (actual > 0).astype(np.int8), (predicted > 0).astype(np.int8)
    metrics['pixel-wise']['iou'].append(iou(actual_bin, predicted_bin))
    metrics['pixel-wise']['dice'].append(dice_coefficient(actual_bin, predicted_bin))
    precision, recall, specif, f1_score_metric = precision_recall_specificity_f1_score_pixel_score(actual_bin, predicted_bin)
    metrics['pixel-wise']['precision'].append(precision), metrics['pixel-wise']['recall'].append(recall)
    metrics['pixel-wise']['specificity'].append(specif), metrics['pixel-wise']['f1-score'].append(f1_score_metric)

    return metrics


def iou(actual, predicted) -> float:
    """Calculates the Intersection over Union (IoU) score."""
    tp = np.sum((actual == 1) & (predicted == 1))
    fp = np.sum((actual == 0) & (predicted == 1))
    fn = np.sum((actual == 1) & (predicted == 0))
    return tp / (tp + fp + fn + EPS)


def dice_coefficient(actual, predicted) -> float:
    """Calculates the Dice Coefficient."""
    tp = np.sum((actual == 1) & (predicted == 1))
    fp = np.sum((actual == 0) & (predicted == 1))
    fn = np.sum((actual == 1) & (predicted == 0))
    return 2 * tp / (2 * tp + fp + fn + EPS)


def matched_segments(actual, predicted, thr=0.5) -> Dict:
    matches = {}
    for label in np.unique(actual)[1:]:
        actual_mask = (actual == label)
        iou_scores = {pred_label: iou(actual_mask, (predicted == pred_label)) for pred_label in
                      np.unique(predicted)[1:]}
        matches[label] = [(pred_label, score) for pred_label, score in iou_scores.items() if score > thr]
    return matches


def panoptic_quality(actual, predicted, matches) -> (float, float, float):
    tp, fp, fn = 0, 0, 0

    matched_actual = set()
    matched_predicted = set()

    for actual_label, preds in matches.items():
        if preds:
            best_match = max(preds, key=lambda x: x[1])  # Choose the best IoU match
            matched_actual.add(actual_label)
            matched_predicted.add(best_match[0])
            tp += 1

    fp = len(np.unique(predicted)[1:]) - len(matched_predicted)
    fn = len(np.unique(actual)[1:]) - len(matched_actual)

    dq = tp / (tp + 0.5 * fp + 0.5 * fn + EPS)

    sq = 0
    for actual_label, preds in matches.items():
        if preds:
            best_match = max(preds, key=lambda x: x[1])
            sq += best_match[1]

    sq /= (tp + EPS)

    return dq * sq, dq, sq


def precision_recall_specificity_f1_score_pixel_score(actual, predicted) -> (float, float, float, float):
    tp = np.sum((actual == 1) & (predicted == 1))
    fp = np.sum((actual == 0) & (predicted == 1))
    fn = np.sum((actual == 1) & (predicted == 0))
    tn = np.sum((actual == 0) & (predicted == 0))

    precision = tp / (tp + fp + EPS)
    recall = tp / (tp + fn + EPS)
    specificity = tn / (tn + fp + EPS)
    f1_score = 2 * precision * recall / (precision + recall + EPS)

    return precision, recall, specificity, f1_score


def mask_to_coords_conversion(mask):
    z, y, x = np.where(mask > 0)
    coords = list(zip(z, y, x))
    return np.array(coords)


def precision_recall_f1_score_coordinates(actual, predicted, threshold=3) -> (float, float, float):
    differences = np.abs(actual[:, None] - predicted)
    below_threshold = np.all(differences < threshold, axis=-1)

    tp = np.sum(np.any(below_threshold, axis=1))
    fp = np.sum(np.logical_not(np.any(below_threshold, axis=0)))
    fn = np.sum(np.logical_not(np.any(below_threshold, axis=1)))

    precision = tp / (tp + fp) if (tp + fp) > 0 else 0
    recall = tp / (tp + fn) if (tp + fn) > 0 else 0
    f1_score = 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0

    return precision, recall, f1_score


def precision_recall_f1_score(matches: Dict[int, List[Tuple[int, float]]]) -> Tuple[float, float, float]:
    tp, fp, fn = 0, 0, 0

    for match in matches.values():
        if match:  # If match is not empty
            for actual_match, iou_score in match:
                if iou_score >= 0.5:  # Assuming IoU threshold of 0.5 for a positive match
                    tp += 1
                else:
                    fp += 1
        else:
            fn += 1  # If there are no matches, it means a false negative

    precision = tp / (tp + fp + EPS)
    recall = tp / (tp + fn + EPS)
    f1_score = 2 * precision * recall / (precision + recall + EPS)

    return precision, recall, f1_score


def fast_bin_auc(actual, predicted, partial=False):
    actual, predicted = actual.flatten(), predicted.flatten()
    if partial:
        n_nonzeros = np.count_nonzero(actual)
        n_zeros = len(actual) - n_nonzeros
        k = min(n_zeros, n_nonzeros)
        predicted = np.concatenate([
            np.sort(predicted[actual == 0])[::-1][:k],
            np.sort(predicted[actual == 1])[::-1][:k]
        ])
        actual = np.concatenate([np.zeros(k), np.ones(k)])

    r = rankdata(predicted)
    n_pos = np.sum(actual)
    n_neg = len(actual) - n_pos
    if n_pos == 0 or n_neg == 0: return 0
    return (np.sum(r[actual == 1]) - n_pos * (n_pos + 1) / 2) / (n_pos * n_neg)


def fast_bin_dice(actual, predicted):
    actual = np.asarray(actual).astype(bool) # bools are way faster to deal with by numpy
    predicted = np.asarray(predicted).astype(bool)
    im_sum = actual.sum() + predicted.sum()
    if im_sum == 0: return 1
    intersection = np.logical_and(actual, predicted)
    return 2. * intersection.sum() / im_sum


"""fluospotter for peak detection and nuclei segmentation.

Modules are arranged as follows:
- augment: Data augmentation to artificially increase dataset size.
- data: Data manipulation. Mainly to properly format for training.
- datasets: Unique data import functions.
- inference: Prediction related functions.
- io: File-manipulation-related functions.
- losses: Simple functions returning model losses.
- metrics: Quantitative output of training / model performance.
- models: Training loop containing classes for each type of model.
- networks: Architecture / building of model structure.
- optimizers: Simple functions returning model optimizers.
- training: Core training loop and callbacks.
- util: Basic utility functions not fitting into a category.
"""

__version__ = "0.1.0"

from . import augment
from . import data
from . import datasets
from . import inference
from . import io
from . import losses
from . import metrics
from . import models
from . import networks
from . import optimizers
from . import training
from . import util



"""Model prediction / inference functions."""
import pdb
import numpy as np
from tqdm import trange
from monai.inferers import sliding_window_inference
from .metrics import compute_segmentation_metrics, compute_puncta_metrics
from .data import match_labeling, join_connected_puncta
from .metrics import fast_bin_auc, fast_bin_dice
from skimage.measure import label
from sklearn.neighbors import KNeighborsClassifier
from skimage.morphology import binary_opening
import time


def validate(model, loader, loss_fn, slwin_bs=2):
    model.eval()
    device = 'cuda' if next(model.parameters()).is_cuda else 'cpu'
    patch_size = model.patch_size
    dscs, aucs, losses = [], [], []
    with trange(len(loader)) as t:
        n_elems, running_dsc = 0, 0
        for val_data in loader:
            images, labels = val_data["img"].to(device), val_data["seg"]
            n_classes = labels.shape[1]
            preds = sliding_window_inference(images, patch_size, slwin_bs, model, overlap=0.1, mode='gaussian').cpu()
            del images
            loss = loss_fn(preds, labels)
            preds = preds.argmax(dim=1).squeeze().numpy()
            labels = labels.squeeze().numpy().astype(np.int8)

            dsc_score, auc_score = [], []
            for l in range(1,n_classes):
                dsc_score.append(fast_bin_dice(labels[l], preds == l))
                auc_score.append(fast_bin_auc(labels[l], preds == l, partial=True))
                if np.isnan(dsc_score[l-1]): dsc_score[l] = 0

            dscs.append(dsc_score)
            aucs.append(auc_score)
            losses.append(loss.item())
            n_elems += 1
            running_dsc += np.mean(dsc_score)
            run_dsc = running_dsc / n_elems
            t.set_postfix(DSC="{:.2f}".format(100 * run_dsc))
            t.update()

    return [100 * np.mean(np.array(dscs)), 100 * np.mean(np.array(aucs)), np.mean(np.array(losses))]


def evaluate(model, loader, slwin_bs=2, compute_metrics=False):
    if model.refinement:
        model_refine = model.refinement
    cfg = model.cfg
    model = model.network
    model.eval()
    device = 'cuda' if next(model.parameters()).is_cuda else 'cpu'
    patch_size = tuple(map(int, cfg["patch_size"].split('/')))
    out = {}
    with trange(len(loader)) as t:
        for val_data in loader:
            start_time = time.time()
            images = val_data["img"].to(device)
            if 'seg' in val_data:
                labels = val_data["seg"]
            else:
                labels = None
            preds = sliding_window_inference(images, patch_size, slwin_bs, model, overlap=0.1, mode='gaussian').cpu()
            preds = preds.argmax(dim=1).squeeze().numpy().astype(np.int8)
            if labels is not None: labels = labels.squeeze().numpy().astype(np.int8)
            if str(cfg["model_type"]) == "puncta_detection":
                preds = join_connected_puncta(images.cpu().detach().numpy()[0][0], label(preds, connectivity=3))
                if labels is not None: labels = join_connected_puncta(images.cpu().detach().numpy()[0][0], label(labels[1], connectivity=3))
            if labels is not None:
                if labels.shape[0] != preds.shape[0]: labels = labels.argmax(0)
            if bool(cfg["instance_seg"]):
                preds = (preds == 2).astype(int)
                preds = binary_opening(preds)
                preds = label(preds, connectivity=3)
                if bool(cfg["refinement"]):
                    binary_mask = sliding_window_inference(images, patch_size, slwin_bs, model_refine, overlap=0.1, mode='gaussian').cpu()
                    binary_mask = binary_mask.argmax(dim=1).squeeze().numpy().astype(np.int8)
                    preds = train_knn(binary_mask, preds)
                    del binary_mask
            print("Elapsed time:", time.time() - start_time, "seconds")
            if compute_metrics:
                if str(cfg["model_type"]) == "segmentation" and labels is not None:
                    preds = match_labeling(labels, preds)
                    out = compute_segmentation_metrics(preds, labels, out)
                elif str(cfg["model_type"]) == "puncta_detection" and labels is not None:
                    out = compute_puncta_metrics(preds, labels, out)
            else:
                if len(out) == 0: out['seg'] = []
                out['seg'].append(preds)
            del images
            del preds
            del labels
            t.update()

    return out


def train_knn(borders: np.array, preds: np.array) -> np.array:
    training_samples = []
    borders = borders.astype(bool) & (~preds.astype(bool))
    borders = borders.astype(int)
    preds[borders == 1] = -1
    for l in np.unique(preds)[1:]:
        z_coords, y_coords, x_coords = np.where(preds == l)
        training_samples.append(np.stack([x_coords, y_coords, z_coords, l * np.ones_like(x_coords)], axis=1))

    training_samples = np.vstack(training_samples)
    x_train, y_train = training_samples[:, :3], training_samples[:, 3]

    knn = KNeighborsClassifier(n_neighbors=1)
    knn.fit(x_train, y_train)

    z_coords, y_coords, x_coords = np.where(borders == 1)
    x_test = np.stack([x_coords, y_coords, z_coords], axis=1)
    y_pred = knn.predict(x_test)

    preds[borders == 1] = 0
    borders[borders == 1] = y_pred.astype(int)

    preds = preds + borders
    return preds


"""puncta_detection class."""

import pdb

from ..datasets import Dataset
from ._models import Model
from ..networks.unet import CustomUNet
from ..training import train_model
from ..inference import evaluate
from ..io import check_puncta_configuration_file, save_metrics_csv
from ..data import get_loaders_test


class SpotsModel(Model):
    """Class to predict spot localization; see base class."""

    def __init__(self, pretrained=None, model_name='small_unet_3d',
                 configuration={}, **kwargs):
        super().__init__(**kwargs)
        self.cfg = check_puncta_configuration_file(configuration)
        self.cfg["model_type"] = "puncta_detection"
        self.network = CustomUNet(model_name=model_name, pretrained=pretrained, in_c=int(self.cfg["in_channels"]), n_classes=2,
                                  patch_size=tuple(map(int, self.cfg["patch_size"].split('/')))).model
        self.model_name = model_name
        self.refinement = None
        self.pretrained = pretrained

    def train(self, dataset: Dataset, **kwargs) -> None:
        if not dataset.training:
            raise ValueError('Dataset does not contain training data.')

        train_model(dataset=dataset, model=self)

    def predict(self, dataset: Dataset) -> None:
        test_loaders = get_loaders_test(data_path=dataset.data_dir, labels_path=dataset.spots_dir,
                                        n_samples=int(self.cfg["n_samples"]), neg_samples=int(self.cfg["neg_samples"]),
                                        patch_size=tuple(map(int, self.cfg["patch_size"].split('/'))),
                                        num_workers=int(self.cfg["num_workers"]),
                                        depth_last=bool(self.cfg["depth_last"]), n_classes=2)
        evaluate(self, test_loaders, compute_metrics=False)

    def evaluate(self, dataset: Dataset, display: bool = True) -> None:
        test_loaders = get_loaders_test(data_path=dataset.data_dir, labels_path=dataset.spots_dir,
                                        n_samples=int(self.cfg["n_samples"]), neg_samples=int(self.cfg["neg_samples"]),
                                        patch_size=tuple(map(int, self.cfg["patch_size"].split('/'))),
                                        num_workers=int(self.cfg["num_workers"]),
                                        depth_last=bool(self.cfg["depth_last"]), n_classes=2, im_size=tuple(map(int, self.cfg["im_size"].split('/'))), instance_seg=bool(self.cfg["instance_seg"]))
        metrics = evaluate(self, test_loaders, compute_metrics=True)
        save_metrics_csv(self.pretrained, metrics)


"""Optimizers are used to update weight parameters in a neural network.

This file contains functions to return standard or custom optimizers.
"""

from typing import Callable, Any, Dict
import torch.optim as optim
import torch.optim.lr_scheduler as lr_scheduler


def adam(learning_rate: float) -> Callable[[Any], optim.Adam]:
    """PyTorch's Adam optimizer with a specified learning rate.

    Args:
        learning_rate (float): The learning rate for the optimizer.

    Returns:
        Callable[[Any], optim.Adam]: A function that takes model parameters and returns an Adam optimizer.
    """
    return lambda params: optim.Adam(params, lr=learning_rate)


def sgd(learning_rate: float, momentum: float = 0.0, weight_decay: float = 0.0) -> Callable[[Any], optim.SGD]:
    """PyTorch's SGD optimizer with a specified learning rate and momentum.

    Args:
        learning_rate (float): The learning rate for the optimizer.
        momentum (float): The momentum factor.
        weight_decay (float): The weight decay factor.

    Returns:
        Callable[[Any], optim.SGD]: A function that takes model parameters and returns an SGD optimizer.
    """
    return lambda params: optim.SGD(params, lr=learning_rate, momentum=momentum, weight_decay=weight_decay)


def cosineAnnealingWarmRestarts(optimizer: optim.Optimizer, T_0: Any, eta_min: Any) -> lr_scheduler.CosineAnnealingWarmRestarts:
    """Cosine Annealing Warm Restarts scheduler.

    Args:
        optimizer (optim.Optimizer): Wrapped optimizer.
        T_0 (int): Number of iterations for the first restart.
        eta_min (float): Minimum learning rate. Default: 0.

    Returns:
        lr_scheduler.CosineAnnealingWarmRestarts: Cosine annealing scheduler with warm restarts.
    """
    return lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=T_0, eta_min=eta_min)


def cosineAnnealingLR(optimizer: optim.Optimizer, T_max: Any, eta_min: Any) -> lr_scheduler.CosineAnnealingLR:
    """Cosine Annealing Learning Rate scheduler.

    Args:
        optimizer (optim.Optimizer): Wrapped optimizer.
        T_max (int): Maximum number of iterations.
        eta_min (float): Minimum learning rate. Default: 0.

    Returns:
        lr_scheduler.CosineAnnealingLR: Cosine annealing learning rate scheduler.
    """
    return lr_scheduler.CosineAnnealingLR(optimizer, T_max=T_max, eta_min=eta_min)


def get_optimizer(optimizer_name: str, optimizer_params: Dict[str, Any], params: Any) -> optim.Optimizer:
    """Get an optimizer by name.

    Args:
        optimizer_name (str): The name of the optimizer.
        optimizer_params (Dict[str, Any]): Parameters for the optimizer.
        params (Any): Model parameters to be optimized.

    Returns:
        optim.Optimizer: The optimizer instance.
    """
    if optimizer_name == 'adam':
        return adam(learning_rate=float(optimizer_params["learning_rate"]))(params)
    elif optimizer_name == 'sgd':
        return sgd(learning_rate=float(optimizer_params["learning_rate"]), momentum=float(optimizer_params.get("momentum", 0.0)), weight_decay=float(optimizer_params.get("weight_decay", 0.0)))(params)
    else:
        raise ValueError(f'Invalid optimizer name: {optimizer_name}')


def get_scheduler(scheduler: str, optimizer: optim.Optimizer, T: Any, eta_min: int = 0) -> lr_scheduler.LRScheduler:
    """Get a learning rate scheduler by name.

    Args:
        scheduler (str): The name of the scheduler.
        optimizer (optim.Optimizer): The optimizer instance.
        T (int): Parameter for the scheduler.
        eta_min (float): Minimum learning rate.

    Returns:
        lr_scheduler._LRScheduler: The learning rate scheduler instance.
    """
    if scheduler == 'cosineAnnealingWarmRestarts':
        return cosineAnnealingWarmRestarts(optimizer, T, eta_min)
    elif scheduler == 'cosineAnnealingLR':
        return cosineAnnealingLR(optimizer, T, eta_min)
    else:
        raise ValueError(f'Invalid scheduler name: {scheduler}')


"""Dataset preparation functions."""

from typing import List, Tuple, Union, Dict, Any
import os
from tifffile import imread
import pandas as pd

# List of currently supported image file extensions.
EXTENSIONS = ("tif", "tiff")


def check_puncta_configuration_file(cfg: Dict) -> Dict:
    params = [
        "model_name", "pretrained", "loss",
        "alpha", "batch_size", "acc_grad", "n_samples", "neg_samples",
        "ovft_check", "patch_size", "optimizer", "lr", "n_epochs", "vl_interval",
        "cyclical_lr", "metric", "num_workers", "depth_last", "in_channels", "detect_overlapping", "full_resolution"
    ]
    values = [
        "small_unet_3d", False, "cedice",
        1.0, 1, 4, 12, 1,
        4, "48/256/256", "adam", 3e-4, 20, 5,
        True, "DSC", 0, False, 1, False, False]

    for p in range(len(params)):
        if not (params[p] in cfg):
            cfg[params[p]] = values[p]
    return cfg


def check_seg_configuration_file(cfg: Dict) -> Dict:
    """Ensures the segmentation configuration dictionary contains all necessary parameters.

        Args:
            cfg: Configuration dictionary.

        Returns:
            Updated configuration dictionary with default values added for missing parameters.
    """
    params = [
        "n_classes", "model_name", "pretrained", "loss1", "loss2",
        "alpha1", "alpha2", "batch_size", "acc_grad", "n_samples", "neg_samples",
        "ovft_check", "patch_size", "optimizer", "lr", "n_epochs", "vl_interval",
        "cyclical_lr", "metric", "num_workers", "depth_last", "in_channels", "refinement", "full_resolution", "instance_seg"
    ]
    values = [
        2, "small_unet_3d", False, "ce", "dice",
        1.0, 1.0, 1, 4, 12, 1,
        4, "48/256/256", "adam", 3e-4, 20, 5,
        True, "DSC", 0, False, 1, False, False, False]

    for p in range(len(params)):
        if not (params[p] in cfg):
            cfg[params[p]] = values[p]
    return cfg


def save_metrics_csv(path: str, metrics: Dict):
    df = pd.DataFrame(metrics)
    dir_path = os.path.dirname(path)
    filename = os.path.basename(os.path.splitext(path)[0]) + ".csv"
    df.to_csv(os.path.join(dir_path, filename))


def load_metrics_csv(path: str) -> dict:
    df = pd.read_csv(path)
    metrics = df.to_dict(orient='list')
    return metrics


def load_files(fname: str, training: bool = False) -> Dict[str, List[str]]:
    """Imports data for custom training and inference.

    Args:
        fname: Path to data files.
        training: Only return testing images and labels if false.

    Returns:
        A dictionary with keys 'train', 'valid', 'test' containing lists of file paths.

    Raises:
        ValueError: If not all datasets are found when training.
    """
    expected = ["train", "valid", "test"]
    check_if_folders_exist = {folder: os.path.isdir(os.path.join(fname, folder)) for folder in expected}

    data = {}

    if training:
        if all(check_if_folders_exist.values()):
            # Training
            for folder in expected:
                data[folder] = load_folder(os.path.join(fname, folder))
        else:
            missing_folders = [folder for folder, exists in check_if_folders_exist.items() if not exists]
            raise ValueError(f"{expected} must be present when training. Missing folders {missing_folders}.")
    else:
        # Inference
        for folder in ["test"]:
            if check_if_folders_exist["test"]:
                data["test"] = load_folder(os.path.join(fname, "test"))
            else:
                raise ValueError("Test folder must be present for inference.")

    return data


def load_folder(fname: str) -> List[str]:
    """Loads all image files from a given folder and checks their consistency in size.

        Args:
            fname: Path to the folder containing image files.

        Returns:
            A list of file paths.

        Raises:
            ValueError: If the files do not coincide in shape or cannot be loaded.
    """
    files = [os.path.join(fname, file) for file in os.listdir(fname) if file.lower().endswith(EXTENSIONS)]
    size = None
    for f in files:
        try:
            file = imread(f)
            if size is None:
                size = file.shape
            elif size != file.shape:
                raise ValueError(f"Files do not coincide in shape. {f} has size {file.shape}, expected {size}")
        except Exception as e:
            raise ValueError(f"Could not load {os.path.join(fname, f)}.") from e
    return files


def validate_data_and_labels(data: Dict[str, List[str]], labels: Dict[str, List[str]], training: bool = True) -> bool:
    """Validates that data and label files match in their basename (excluding extensions).

        Args:
            data: Dictionary containing lists of data file paths.
            labels: Dictionary containing lists of label file paths.

        Returns:
            True if data and label files match, False otherwise.
    """
    if training:
        keys = ['train', 'valid', 'test']
    else:
        keys = ['test']
    for key in keys:
        if remove_extension(data[key]) != remove_extension(labels[key]):
            return False
    return True


def remove_extension(file_list):
    """Removes the file extension from a list of file paths.

        Args:
            file_list: List of file paths.

        Returns:
            A set of file basenames without extensions.
    """
    return {os.path.basename(os.path.splitext(file)[0]) for file in file_list}



"""Models module with the training loop and logic to handle data which feeds into the loop."""

from ._models import Model
from .nuclei import SegmentationModel
from .puncta import SpotsModel

__all__ = ["Model", "segmentation", "puncta_detection"]


from setuptools import setup, find_packages

setup(
    name='fluospotter',
    version='0.1',
    packages=find_packages(),
    license="CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
    install_requires=[
        'numpy',
        'torch',
        'monai',
        'scikit-image',
        'sklearn',
        'tifffile',
        'pandas'
    ],
    author='Arnau Blanco Borrego',
    author_email='blancoarnau@gmail.com',
    description='Python library designed for nuclei segmentation and puncta detection in fluorescence microscopy images.',
    url='https://github.com/arnaublanco/Fluospotter',
)


"""UNet architecture."""
import pdb

import torch
from monai.networks.nets import UNet, DynUNet


class CustomUNet:
    def __init__(self, model_name, pretrained=None, in_c=1, n_classes=2, patch_size=(64, 64, 16)):
        self.model = self.get_model(model_name=model_name, in_c=in_c, n_classes=n_classes, pretrained=pretrained,
                                    patch_size=patch_size)

    @staticmethod
    def get_kernels_strides(patch_size=(64, 64, 16), spacings=(1.5625, 1.5625, 5.0)):
        input_size = patch_size
        strides, kernels = [], []
        while True:
            spacing_ratio = [sp / min(spacings) for sp in spacings]
            stride = [2 if ratio <= 2 and size >= 8 else 1 for (ratio, size) in zip(spacing_ratio, patch_size)]
            kernel = [3 if ratio <= 2 else 1 for ratio in spacing_ratio]
            if all(s == 1 for s in stride):
                break
            for idx, (i, j) in enumerate(zip(patch_size, stride)):
                if i % j != 0:
                    raise ValueError(
                        f"Patch size is not supported, please try to modify the size {input_size[idx]} in the spatial dimension {idx}."
                    )
            patch_size = [i / j for i, j in zip(patch_size, stride)]
            spacings = [i * j for i, j in zip(spacings, stride)]
            kernels.append(kernel)
            strides.append(stride)

        strides.insert(0, len(spacings) * [1])
        kernels.append(len(spacings) * [3])
        return kernels, strides

    @staticmethod
    def get_dynunet(patch_size=(64, 64, 16), spacings=(1.5625, 1.5625, 5.0), in_channels=1, n_classes=2,
                    deep_supr_num=0):
        kernels, strides = CustomUNet.get_kernels_strides(patch_size, spacings)
        deep_supervision = False if deep_supr_num == 0 else True

        net = DynUNet(spatial_dims=3, in_channels=in_channels, out_channels=n_classes, kernel_size=kernels,
                      strides=strides,
                      upsample_kernel_size=strides[1:], norm_name="instance", deep_supervision=deep_supervision,
                      deep_supr_num=deep_supr_num)

        return net

    @staticmethod
    def get_model(model_name, in_c=1, n_classes=2, pretrained=None, patch_size=(64, 64, 16)):

        if model_name == 'small_unet_3d':
            model = UNet(spatial_dims=3, in_channels=in_c, out_channels=n_classes, channels=(16, 32,), strides=(2,),
                         num_res_units=1)
            if pretrained is not None:
                try:
                    model.load_state_dict(torch.load(pretrained))
                except Exception:
                    raise ValueError(f'Could not load pretrained weights {pretrained} for small_unet_3d.')
        elif model_name == 'dynunet':
            # if patch_size = (256,256,48) -> spacings (1., 1., 5.)
            # if patch_size = (48,256,256) -> spacings (5., 1., 1.)
            model = CustomUNet.get_dynunet(patch_size=patch_size, spacings=(5., 1., 1.), in_channels=in_c, n_classes=n_classes)
            if pretrained is not None:
                try:
                    model.load_state_dict(torch.load(pretrained))
                except Exception:
                    raise ValueError(f'Could not load pretrained weights {pretrained} for dynunet.')
        else:
            raise ValueError(f'The given nuclei segmentation model \'{model_name}\' is not valid.')

        setattr(model, 'n_classes', n_classes)
        setattr(model, 'patch_size', patch_size)

        return model